@@ -17,26 +17,6 @@
 #include "storage/shm_mq.h"
 #include "utils/guc_tables.h"
 
-static inline TupleDesc
-CreateTemplateTupleDescCompat(int nattrs, bool hasoid)
-{
-#if PG_VERSION_NUM >= 120000
-	return CreateTemplateTupleDesc(nattrs);
-#else
-	return CreateTemplateTupleDesc(nattrs, hasoid);
-#endif
-}
-
-static inline void
-shm_mq_detach_compat(shm_mq_handle *mqh, shm_mq *mq)
-{
-#if PG_VERSION_NUM >= 100000
-	shm_mq_detach(mqh);
-#else
-	shm_mq_detach(mq);
-#endif
-}
-
 static inline shm_mq_result
 shm_mq_send_compat(shm_mq_handle *mqh, Size nbytes, const void *data,
 				   bool nowait, bool force_flush)
@@ -48,24 +28,26 @@ shm_mq_send_compat(shm_mq_handle *mqh, Size nbytes, const void *data,
 #endif
 }
 
+#if PG_VERSION_NUM < 170000
+#define INIT_PG_LOAD_SESSION_LIBS		0x0001
+#define INIT_PG_OVERRIDE_ALLOW_CONNS	0x0002
+#endif
+
 static inline void
 InitPostgresCompat(const char *in_dbname, Oid dboid,
 				   const char *username, Oid useroid,
-				   bool load_session_libraries,
-				   bool override_allow_connections,
+				   bits32 flags,
 				   char *out_dbname)
 {
 #if PG_VERSION_NUM >= 170000
-	InitPostgres(in_dbname, dboid, username, useroid, (load_session_libraries ? INIT_PG_LOAD_SESSION_LIBS : 0) |
-				 (override_allow_connections ? INIT_PG_OVERRIDE_ALLOW_CONNS : 0), out_dbname);
+	InitPostgres(in_dbname, dboid, username, useroid, flags, out_dbname);
 #elif PG_VERSION_NUM >= 150000
-	InitPostgres(in_dbname, dboid, username, useroid, load_session_libraries,
-				 override_allow_connections, out_dbname);
-#elif PG_VERSION_NUM >= 110000
-	InitPostgres(in_dbname, dboid, username, useroid, out_dbname,
-				 override_allow_connections);
+	InitPostgres(in_dbname, dboid, username, useroid,
+				 flags & INIT_PG_LOAD_SESSION_LIBS,
+				 flags & INIT_PG_OVERRIDE_ALLOW_CONNS, out_dbname);
 #else
-	InitPostgres(in_dbname, dboid, username, useroid, out_dbname);
+	InitPostgres(in_dbname, dboid, username, useroid, out_dbname,
+				 flags & INIT_PG_OVERRIDE_ALLOW_CONNS);
 #endif
 }
 
