@@ -18,15 +18,14 @@
 #include "optimizer/planner.h"
 #include "pgstat.h"
 #include "postmaster/autovacuum.h"
-#if PG_VERSION_NUM >= 120000
 #include "replication/walsender.h"
-#endif
 #include "storage/ipc.h"
 #include "storage/pg_shmem.h"
 #include "storage/procarray.h"
 #include "storage/shm_mq.h"
 #include "storage/shm_toc.h"
 #include "storage/spin.h"
+#include "tcop/utility.h"
 #include "utils/builtins.h"
 #include "utils/datetime.h"
 #include "utils/guc_tables.h"
@@ -44,8 +43,15 @@ void		_PG_init(void);
 static bool shmem_initialized = false;
 
 /* Hooks */
+static ExecutorStart_hook_type	prev_ExecutorStart = NULL;
+static ExecutorRun_hook_type 	prev_ExecutorRun = NULL;
+static ExecutorFinish_hook_type prev_ExecutorFinish = NULL;
 static ExecutorEnd_hook_type	prev_ExecutorEnd = NULL;
 static planner_hook_type		planner_hook_next = NULL;
+static ProcessUtility_hook_type prev_ProcessUtility = NULL;
+
+/* Current nesting depth of planner/Executor calls */
+static int nesting_level = 0;
 
 /* Pointers to shared memory objects */
 shm_mq				   *pgws_collector_mq = NULL;
@@ -67,7 +73,56 @@ static PlannedStmt *pgws_planner_hook(Query *parse,
 		const char *query_string,
 #endif
 		int cursorOptions, ParamListInfo boundParams);
+static void pgws_ExecutorStart(QueryDesc *queryDesc, int eflags);
+static void pgws_ExecutorRun(QueryDesc *queryDesc,
+							 ScanDirection direction,
+							 uint64 count, bool execute_once);
+static void pgws_ExecutorFinish(QueryDesc *queryDesc);
 static void pgws_ExecutorEnd(QueryDesc *queryDesc);
+static void pgws_ProcessUtility(PlannedStmt *pstmt,
+								const char *queryString,
+#if PG_VERSION_NUM >= 140000
+								bool readOnlyTree,
+#endif
+								ProcessUtilityContext context,
+								ParamListInfo params,
+								QueryEnvironment *queryEnv,
+								DestReceiver *dest,
+#if PG_VERSION_NUM >= 130000
+								QueryCompletion *qc
+#else
+								char *completionTag
+#endif
+								);
+
+/*---- GUC variables ----*/
+
+typedef enum
+{
+	PGWS_PROFILE_QUERIES_NONE,			/* profile no statements */
+	PGWS_PROFILE_QUERIES_TOP,			/* only top level statements */
+	PGWS_PROFILE_QUERIES_ALL			/* all statements, including nested ones */
+} PGWSTrackLevel;
+
+static const struct config_enum_entry pgws_profile_queries_options[] =
+{
+	{"none", 	PGWS_PROFILE_QUERIES_NONE, false},
+	{"off", 	PGWS_PROFILE_QUERIES_NONE, false},
+	{"no", 		PGWS_PROFILE_QUERIES_NONE, false},
+	{"false", 	PGWS_PROFILE_QUERIES_NONE, false},
+	{"0", 		PGWS_PROFILE_QUERIES_NONE, false},
+	{"top", 	PGWS_PROFILE_QUERIES_TOP, false},
+	{"on", 		PGWS_PROFILE_QUERIES_TOP, false},
+	{"yes", 	PGWS_PROFILE_QUERIES_TOP, false},
+	{"true", 	PGWS_PROFILE_QUERIES_TOP, false},
+	{"1", 		PGWS_PROFILE_QUERIES_TOP, false},
+	{"all", 	PGWS_PROFILE_QUERIES_ALL, false},
+	{NULL, 		0, false}
+};
+
+#define pgws_enabled(level) \
+	((pgws_collector_hdr->profileQueries == PGWS_PROFILE_QUERIES_ALL) || \
+	 (pgws_collector_hdr->profileQueries == PGWS_PROFILE_QUERIES_TOP && (level) == 0))
 
 /*
  * Calculate max processes count.
@@ -117,9 +172,7 @@ get_max_procs_count(void)
 	 * Starting with pg12, wal senders aren't part of MaxConnections anymore
 	 * and have to be accounted for.
 	 */
-#if PG_VERSION_NUM >= 120000
 	count += max_wal_senders;
-#endif		/* pg 12+ */
 #endif		/* pg 15- */
 	/* End of MaxBackends calculation. */
 
@@ -161,6 +214,14 @@ shmem_int_guc_check_hook(int *newval, void **extra, GucSource source)
 	return true;
 }
 
+static bool
+shmem_enum_guc_check_hook(int *newval, void **extra, GucSource source)
+{
+	if (UsedShmemSegAddr == NULL)
+		return false;
+	return true;
+}
+
 static bool
 shmem_bool_guc_check_hook(bool *newval, void **extra, GucSource source)
 {
@@ -196,7 +257,8 @@ setup_gucs()
 				history_period_found = false,
 				profile_period_found = false,
 				profile_pid_found = false,
-				profile_queries_found = false;
+				profile_queries_found = false,
+				sample_cpu_found = false;
 
 	get_guc_variables_compat(&guc_vars, &numOpts);
 
@@ -235,8 +297,14 @@ setup_gucs()
 		else if (!strcmp(name, "pg_wait_sampling.profile_queries"))
 		{
 			profile_queries_found = true;
-			var->_bool.variable = &pgws_collector_hdr->profileQueries;
-			pgws_collector_hdr->profileQueries = true;
+			var->_enum.variable = &pgws_collector_hdr->profileQueries;
+			pgws_collector_hdr->profileQueries = PGWS_PROFILE_QUERIES_TOP;
+		}
+		else if (!strcmp(name, "pg_wait_sampling.sample_cpu"))
+		{
+			sample_cpu_found = true;
+			var->_bool.variable = &pgws_collector_hdr->sampleCpu;
+			pgws_collector_hdr->sampleCpu = true;
 		}
 	}
 
@@ -265,16 +333,23 @@ setup_gucs()
 				PGC_SUSET, 0, shmem_bool_guc_check_hook, NULL, NULL);
 
 	if (!profile_queries_found)
-		DefineCustomBoolVariable("pg_wait_sampling.profile_queries",
+		DefineCustomEnumVariable("pg_wait_sampling.profile_queries",
 				"Sets whether profile should be collected per query.", NULL,
-				&pgws_collector_hdr->profileQueries, true,
-				PGC_SUSET, 0, shmem_bool_guc_check_hook, NULL, NULL);
+				&pgws_collector_hdr->profileQueries, PGWS_PROFILE_QUERIES_TOP, pgws_profile_queries_options,
+				PGC_SUSET, 0, shmem_enum_guc_check_hook, NULL, NULL);
+
+	if (!sample_cpu_found)
+		DefineCustomBoolVariable("pg_wait_sampling.sample_cpu",
+		                         "Sets whether not waiting backends should be sampled.", NULL,
+		                         &pgws_collector_hdr->sampleCpu, true,
+		                         PGC_SUSET, 0, shmem_bool_guc_check_hook, NULL, NULL);
 
 	if (history_size_found
 		|| history_period_found
 		|| profile_period_found
 		|| profile_pid_found
-		|| profile_queries_found)
+		|| profile_queries_found
+	    || sample_cpu_found)
 	{
 		ProcessConfigFile(PGC_SIGHUP);
 	}
@@ -316,6 +391,8 @@ pgws_shmem_startup(void)
 
 		pgws_collector_hdr = shm_toc_allocate(toc, sizeof(CollectorShmqHeader));
 		shm_toc_insert(toc, 0, pgws_collector_hdr);
+		/* needed to please check_GUC_init */
+		pgws_collector_hdr->profileQueries = PGWS_PROFILE_QUERIES_TOP;
 		pgws_collector_mq = shm_toc_allocate(toc, COLLECTOR_QUEUE_SIZE);
 		shm_toc_insert(toc, 1, pgws_collector_mq);
 		pgws_proc_queryids = shm_toc_allocate(toc,
@@ -329,16 +406,9 @@ pgws_shmem_startup(void)
 	else
 	{
 		toc = shm_toc_attach(PG_WAIT_SAMPLING_MAGIC, pgws);
-
-#if PG_VERSION_NUM >= 100000
 		pgws_collector_hdr = shm_toc_lookup(toc, 0, false);
 		pgws_collector_mq = shm_toc_lookup(toc, 1, false);
 		pgws_proc_queryids = shm_toc_lookup(toc, 2, false);
-#else
-		pgws_collector_hdr = shm_toc_lookup(toc, 0);
-		pgws_collector_mq = shm_toc_lookup(toc, 1);
-		pgws_proc_queryids = shm_toc_lookup(toc, 2);
-#endif
 	}
 
 	shmem_initialized = true;
@@ -364,7 +434,7 @@ static void
 pgws_cleanup_callback(int code, Datum arg)
 {
 	elog(DEBUG3, "pg_wait_sampling cleanup: detaching shm_mq and releasing queue lock");
-	shm_mq_detach_compat(recv_mqh, recv_mq);
+	shm_mq_detach(recv_mqh);
 	LockRelease(&queueTag, ExclusiveLock, false);
 }
 
@@ -402,8 +472,16 @@ _PG_init(void)
 	shmem_startup_hook		= pgws_shmem_startup;
 	planner_hook_next		= planner_hook;
 	planner_hook			= pgws_planner_hook;
+	prev_ExecutorStart		= ExecutorStart_hook;
+	ExecutorStart_hook		= pgws_ExecutorStart;
+	prev_ExecutorRun		= ExecutorRun_hook;
+	ExecutorRun_hook		= pgws_ExecutorRun;
+	prev_ExecutorFinish		= ExecutorFinish_hook;
+	ExecutorFinish_hook		= pgws_ExecutorFinish;
 	prev_ExecutorEnd		= ExecutorEnd_hook;
 	ExecutorEnd_hook		= pgws_ExecutorEnd;
+	prev_ProcessUtility		= ProcessUtility_hook;
+	ProcessUtility_hook		= pgws_ProcessUtility;
 }
 
 /*
@@ -432,6 +510,28 @@ search_proc(int pid)
 	return NULL;
 }
 
+/*
+ * Decide whether this PGPROC entry should be included in profiles and output
+ * views.
+ */
+bool
+pgws_should_sample_proc(PGPROC *proc)
+{
+	if (proc->wait_event_info == 0 && !pgws_collector_hdr->sampleCpu)
+		return false;
+
+	/*
+	 * On PostgreSQL versions < 17 the PGPROC->pid field is not reset on
+	 * process exit. This would lead to such processes getting counted for
+	 * null wait events. So instead we make use of DisownLatch() resetting
+	 * owner_pid during ProcKill().
+	 */
+	if (proc->pid == 0 || proc->procLatch.owner_pid == 0 || proc->pid == MyProcPid)
+		return false;
+
+	return true;
+}
+
 typedef struct
 {
 	HistoryItem	   *items;
@@ -459,7 +559,7 @@ pg_wait_sampling_get_current(PG_FUNCTION_ARGS)
 		params->ts = GetCurrentTimestamp();
 
 		funcctx->user_fctx = params;
-		tupdesc = CreateTemplateTupleDescCompat(4, false);
+		tupdesc = CreateTemplateTupleDesc(4);
 		TupleDescInitEntry(tupdesc, (AttrNumber) 1, "pid",
 						   INT4OID, -1, 0);
 		TupleDescInitEntry(tupdesc, (AttrNumber) 2, "type",
@@ -497,13 +597,13 @@ pg_wait_sampling_get_current(PG_FUNCTION_ARGS)
 			{
 				PGPROC *proc = &ProcGlobal->allProcs[i];
 
-				if (proc != NULL && proc->pid != 0 && proc->wait_event_info)
-				{
-					params->items[j].pid = proc->pid;
-					params->items[j].wait_event_info = proc->wait_event_info;
-					params->items[j].queryId = pgws_proc_queryids[i];
-					j++;
-				}
+				if (!pgws_should_sample_proc(proc))
+					continue;
+
+				params->items[j].pid = proc->pid;
+				params->items[j].wait_event_info = proc->wait_event_info;
+				params->items[j].queryId = pgws_proc_queryids[i];
+				j++;
 			}
 			funcctx->max_calls = j;
 		}
@@ -647,7 +747,7 @@ receive_array(SHMRequest request, Size item_size, Size *count)
 	PG_END_ENSURE_ERROR_CLEANUP(pgws_cleanup_callback, 0);
 
 	/* We still have to detach and release lock during normal operation. */
-	shm_mq_detach_compat(recv_mqh, recv_mq);
+	shm_mq_detach(recv_mqh);
 	LockRelease(&queueTag, ExclusiveLock, false);
 
 	return result;
@@ -680,7 +780,7 @@ pg_wait_sampling_get_profile(PG_FUNCTION_ARGS)
 		funcctx->max_calls = profile->count;
 
 		/* Make tuple descriptor */
-		tupdesc = CreateTemplateTupleDescCompat(5, false);
+		tupdesc = CreateTemplateTupleDesc(5);
 		TupleDescInitEntry(tupdesc, (AttrNumber) 1, "pid",
 						   INT4OID, -1, 0);
 		TupleDescInitEntry(tupdesc, (AttrNumber) 2, "type",
@@ -797,7 +897,7 @@ pg_wait_sampling_get_history(PG_FUNCTION_ARGS)
 		funcctx->max_calls = history->count;
 
 		/* Make tuple descriptor */
-		tupdesc = CreateTemplateTupleDescCompat(5, false);
+		tupdesc = CreateTemplateTupleDesc(5);
 		TupleDescInitEntry(tupdesc, (AttrNumber) 1, "pid",
 						   INT4OID, -1, 0);
 		TupleDescInitEntry(tupdesc, (AttrNumber) 2, "sample_ts",
@@ -872,38 +972,126 @@ pgws_planner_hook(Query *parse,
 				  int cursorOptions,
 				  ParamListInfo boundParams)
 {
-	if (MyProc)
-	{
-		int i = MyProc - ProcGlobal->allProcs;
-#if PG_VERSION_NUM >= 110000
-		/*
-		 * since we depend on queryId we need to check that its size
-		 * is uint64 as we coded in pg_wait_sampling
-		 */
-		StaticAssertExpr(sizeof(parse->queryId) == sizeof(uint64),
-				"queryId size is not uint64");
-#else
-		StaticAssertExpr(sizeof(parse->queryId) == sizeof(uint32),
-				"queryId size is not uint32");
-#endif
-		if (!pgws_proc_queryids[i])
-			pgws_proc_queryids[i] = parse->queryId;
+	PlannedStmt    *result;
+	int				i = MyProc - ProcGlobal->allProcs;
+	uint64			save_queryId = 0;
 
+	if (pgws_enabled(nesting_level))
+	{
+		save_queryId = pgws_proc_queryids[i];
+		pgws_proc_queryids[i] = parse->queryId;
 	}
 
-	/* Invoke original hook if needed */
-	if (planner_hook_next)
-		return planner_hook_next(parse,
+	nesting_level++;
+	PG_TRY();
+	{
+		/* Invoke original hook if needed */
+		if (planner_hook_next)
+			result = planner_hook_next(parse,
 #if PG_VERSION_NUM >= 130000
-				query_string,
+					query_string,
 #endif
-				cursorOptions, boundParams);
+					cursorOptions, boundParams);
+		else
+			result = standard_planner(parse,
+#if PG_VERSION_NUM >= 130000
+					query_string,
+#endif
+					cursorOptions, boundParams);
+		nesting_level--;
+		if (nesting_level == 0)
+			pgws_proc_queryids[i] = UINT64CONST(0);
+		else if (pgws_enabled(nesting_level))
+			pgws_proc_queryids[i] = save_queryId;
+	}
+	PG_CATCH();
+	{
+		nesting_level--;
+		if (nesting_level == 0)
+			pgws_proc_queryids[i] = UINT64CONST(0);
+		else if (pgws_enabled(nesting_level))
+			pgws_proc_queryids[i] = save_queryId;
+		PG_RE_THROW();
+	}
+	PG_END_TRY();
 
-	return standard_planner(parse,
-#if PG_VERSION_NUM >= 130000
-				query_string,
-#endif
-			cursorOptions, boundParams);
+	return result;
+}
+
+/*
+ * ExecutorStart hook: save queryId for collector
+ */
+static void
+pgws_ExecutorStart(QueryDesc *queryDesc, int eflags)
+{
+	int i = MyProc - ProcGlobal->allProcs;
+	if (pgws_enabled(nesting_level))
+		pgws_proc_queryids[i] = queryDesc->plannedstmt->queryId;
+	if (prev_ExecutorStart)
+		prev_ExecutorStart(queryDesc, eflags);
+	else
+		standard_ExecutorStart(queryDesc, eflags);
+}
+
+static void
+pgws_ExecutorRun(QueryDesc *queryDesc,
+				ScanDirection direction,
+				uint64 count, bool execute_once)
+{
+	int		i = MyProc - ProcGlobal->allProcs;
+	uint64	save_queryId = pgws_proc_queryids[i];
+
+	nesting_level++;
+	PG_TRY();
+	{
+		if (prev_ExecutorRun)
+			prev_ExecutorRun(queryDesc, direction, count, execute_once);
+		else
+			standard_ExecutorRun(queryDesc, direction, count, execute_once);
+		nesting_level--;
+		if (nesting_level == 0)
+			pgws_proc_queryids[i] = UINT64CONST(0);
+		else
+			pgws_proc_queryids[i] = save_queryId;
+	}
+	PG_CATCH();
+	{
+		nesting_level--;
+		if (nesting_level == 0)
+			pgws_proc_queryids[i] = UINT64CONST(0);
+		else
+			pgws_proc_queryids[i] = save_queryId;
+		PG_RE_THROW();
+	}
+	PG_END_TRY();
+}
+
+static void
+pgws_ExecutorFinish(QueryDesc *queryDesc)
+{
+	int		i = MyProc - ProcGlobal->allProcs;
+	uint64	save_queryId = pgws_proc_queryids[i];
+
+	nesting_level++;
+	PG_TRY();
+	{
+		if (prev_ExecutorFinish)
+			prev_ExecutorFinish(queryDesc);
+		else
+			standard_ExecutorFinish(queryDesc);
+		nesting_level--;
+		if (nesting_level == 0)
+			pgws_proc_queryids[i] = UINT64CONST(0);
+		else
+			pgws_proc_queryids[i] = save_queryId;
+	}
+	PG_CATCH();
+	{
+		nesting_level--;
+		pgws_proc_queryids[i] = save_queryId;
+		PG_RE_THROW();
+	}
+	PG_END_TRY();
 }
 
 /*
@@ -912,11 +1100,85 @@ pgws_planner_hook(Query *parse,
 static void
 pgws_ExecutorEnd(QueryDesc *queryDesc)
 {
-	if (MyProc)
-		pgws_proc_queryids[MyProc - ProcGlobal->allProcs] = UINT64CONST(0);
+	int i = MyProc - ProcGlobal->allProcs;
+	if (nesting_level == 0)
+		pgws_proc_queryids[i] = UINT64CONST(0);
 
 	if (prev_ExecutorEnd)
 		prev_ExecutorEnd(queryDesc);
 	else
 		standard_ExecutorEnd(queryDesc);
 }
+
+static void
+pgws_ProcessUtility(PlannedStmt *pstmt,
+					const char *queryString,
+#if PG_VERSION_NUM >= 140000
+					bool readOnlyTree,
+#endif
+					ProcessUtilityContext context,
+					ParamListInfo params,
+					QueryEnvironment *queryEnv,
+					DestReceiver *dest,
+#if PG_VERSION_NUM >= 130000
+					QueryCompletion *qc
+#else
+					char *completionTag
+#endif
+					)
+{
+	int				i = MyProc - ProcGlobal->allProcs;
+	uint64			save_queryId = 0;
+
+	if (pgws_enabled(nesting_level))
+	{
+		save_queryId = pgws_proc_queryids[i];
+		pgws_proc_queryids[i] = pstmt->queryId;
+	}
+
+	nesting_level++;
+	PG_TRY();
+	{
+		if (prev_ProcessUtility)
+			prev_ProcessUtility (pstmt, queryString,
+#if PG_VERSION_NUM >= 140000
+								 readOnlyTree,
+#endif
+								 context, params, queryEnv,
+								 dest,
+#if PG_VERSION_NUM >= 130000
+								 qc
+#else
+								 completionTag
+#endif
+								 );
+		else
+			standard_ProcessUtility(pstmt, queryString,
+#if PG_VERSION_NUM >= 140000
+									readOnlyTree,
+#endif
+									context, params, queryEnv,
+									dest,
+#if PG_VERSION_NUM >= 130000
+									qc
+#else
+									completionTag
+#endif
+									);
+		nesting_level--;
+		if (nesting_level == 0)
+			pgws_proc_queryids[i] = UINT64CONST(0);
+		else if (pgws_enabled(nesting_level))
+			pgws_proc_queryids[i] = save_queryId;
+	}
+	PG_CATCH();
+	{
+		nesting_level--;
+		if (nesting_level == 0)
+			pgws_proc_queryids[i] = UINT64CONST(0);
+		else if (pgws_enabled(nesting_level))
+			pgws_proc_queryids[i] = save_queryId;
+		PG_RE_THROW();
+	}
+	PG_END_TRY();
+}
