@@ -10,12 +10,7 @@
 #ifndef __PG_WAIT_SAMPLING_H__
 #define __PG_WAIT_SAMPLING_H__
 
-#include <postgres.h>
-
-/* Check PostgreSQL version */
-#if PG_VERSION_NUM < 90600
-	#error "You are trying to build pg_wait_sampling with PostgreSQL version lower than 9.6.  Please, check you environment."
-#endif
+#include "postgres.h"
 
 #include "storage/proc.h"
 #include "storage/shm_mq.h"
@@ -67,7 +62,8 @@ typedef struct
 	int				historyPeriod;
 	int				profilePeriod;
 	bool			profilePid;
-	bool			profileQueries;
+	int				profileQueries;
+	bool			sampleCpu;
 } CollectorShmqHeader;
 
 /* pg_wait_sampling.c */
@@ -75,6 +71,7 @@ extern CollectorShmqHeader *pgws_collector_hdr;
 extern shm_mq			   *pgws_collector_mq;
 extern uint64			   *pgws_proc_queryids;
 extern void pgws_init_lock_tag(LOCKTAG *tag, uint32 lock);
+extern bool pgws_should_sample_proc(PGPROC *proc);
 
 /* collector.c */
 extern void pgws_register_wait_collector(void);
